﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace TestProject
{
    /// <summary>
    /// 픽셀 데이터
    /// </summary>
    public class PixelData
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        /// 비트맵 크기
        /// </summary>
        private Size bitmapDataSize;

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region Field

        /// <summary>
        /// 픽셀 카운트
        /// </summary>
        protected readonly int pixelCount;

        /// <summary>
        /// 픽셀 배열
        /// </summary>
        protected readonly int[] pixelArray;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 비트맵 데이터 크기 - BitmapDataSize

        /// <summary>
        /// 비트맵 데이터 크기
        /// </summary>
        public Size BitmapDataSize
        {
            get
            {
                return this.bitmapDataSize;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Indexer
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 인덱서 - this[point]

        /// <summary>
        /// 인덱서
        /// </summary>
        /// <param name="point">좌표</param>
        /// <returns>픽셀 값</returns>
        public int this[Point point]
        {
            get
            {
                int index = (point.Y * this.bitmapDataSize.Width) + point.X;

                return this.pixelArray[index];
            }
            protected set
            {
                int index = (point.Y * this.bitmapDataSize.Width) + point.X;

                this.pixelArray[index] = value;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 생성자 - PixelData(bitmapData)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="bitmapData">비트맵 데이터</param>
        protected PixelData(BitmapData bitmapData)
        {
            this.bitmapDataSize = new Size(bitmapData.Width, bitmapData.Height);

            this.pixelCount = this.bitmapDataSize.Width * this.bitmapDataSize.Height;

            this.pixelArray = new int[this.pixelCount];

            Marshal.Copy(bitmapData.Scan0, this.pixelArray, 0, this.pixelCount);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 생성하기 - Create(bitmap)

        /// <summary>
        /// 생성하기
        /// </summary>
        /// <param name="bitmap">비트맵</param>
        /// <returns>픽셀 데이터</returns>
        public static PixelData Create(Bitmap bitmap)
        {
            BitmapData bitmapData = GetBitmapData(bitmap, true);

            PixelData pixelData = new PixelData(bitmapData);

            bitmap.UnlockBits(bitmapData);

            return pixelData;
        }

        #endregion
        #region 생성하기 - Create(filePath)

        /// <summary>
        /// 생성하기
        /// </summary>
        /// <param name="filePath">파일 경로</param>
        /// <returns>픽셀 데이터</returns>
        public static PixelData Create(string filePath)
        {
            using(Bitmap bitmap = new Bitmap(filePath))
            {
                BitmapData bitmapData = GetBitmapData(bitmap, true);

                PixelData pixelData = new PixelData(bitmapData);

                bitmap.UnlockBits(bitmapData);

                return pixelData;
            }
        }

        #endregion

        #region 포함 여부 구하기 - Contains(point)

        /// <summary>
        /// 포함 여부 구하기
        /// </summary>
        /// <param name="point">좌표</param>
        /// <returns>포함 여부</returns>
        public bool Contains(Point point)
        {
            return ((point.X < 0) || (point.X >= this.bitmapDataSize.Width) || (point.Y < 0) || (point.Y >= this.bitmapDataSize.Height)) ? false : true;
        }

        #endregion
        #region 색상 여부 구하기 - IsColor(point, predicate)

        /// <summary>
        /// 색상 여부 구하기
        /// </summary>
        /// <param name="point">좌표</param>
        /// <param name="predicate">판정 함수</param>
        /// <returns>색상 여부</returns>
        public bool IsColor(Point point, Predicate<int> predicate)
        {
            int pixel = this[point];

            return predicate(pixel);
        }

        #endregion
        #region 투명색 여부 구하기 - IsTransparentColor(argb)

        /// <summary>
        /// 투명색 여부 구하기
        /// </summary>
        /// <param name="argb">ARGB 값</param>
        /// <returns>투명색 여부</returns>
        public bool IsTransparentColor(int argb)
        {
            Color color = Color.FromArgb(argb);

            return (color.A == 0);
        }

        #endregion
        #region 투명색 여부 구하기 - IsTransparentColor(point)

        /// <summary>
        /// 투명색 여부 구하기
        /// </summary>
        /// <param name="point">좌표</param>
        /// <returns>투명색 여부</returns>
        public bool IsTransparentColor(Point point)
        {
            if(!Contains(point))
            {
                return true;
            }

            return IsColor(point, IsTransparentColor);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Protected

        #region 비트맵 데이터 구하기 - GetBitmapData(bitmap, readOnly)

        /// <summary>
        /// 비트맵 데이터 구하기
        /// </summary>
        /// <param name="bitmap">비트맵</param>
        /// <param name="readOnly">읽기 전용 여부</param>
        /// <returns>비트맵 데이터</returns>
        protected static BitmapData GetBitmapData(Bitmap bitmap, bool readOnly)
        {
            Rectangle rectangle = new Rectangle(new Point(0, 0), bitmap.Size);

            return bitmap.LockBits(rectangle, (readOnly) ? ImageLockMode.ReadOnly : ImageLockMode.ReadWrite, PixelFormat.Format32bppArgb);
        }

        #endregion
    }
}